from functools import partial

from .landmark import ljson_exporter, pts_exporter
from .image import pil_exporter
from .video import ffmpeg_video_exporter
from .pickle import pickle_exporter


landmark_types = {
    '.ljson': ljson_exporter,
    '.pts': pts_exporter
}


image_types = {
    '.bmp': pil_exporter,
    '.dib': pil_exporter,
    '.dcx': pil_exporter,
    '.eps': pil_exporter,
    '.ps': pil_exporter,
    '.gif': pil_exporter,
    '.im': pil_exporter,
    '.jpg': pil_exporter,
    '.jpe': pil_exporter,
    '.jpeg': pil_exporter,
    '.pcd': pil_exporter,
    '.pcx': pil_exporter,
    '.png': pil_exporter,
    '.pbm': pil_exporter,
    '.pgm': pil_exporter,
    '.ppm': pil_exporter,
    '.psd': pil_exporter,
    '.tif': pil_exporter,
    '.tiff': pil_exporter,
    '.xbm': pil_exporter,
    '.xpm': pil_exporter
}


pickle_types = {
    '.pkl': pickle_exporter,
    '.pkl.gz': pickle_exporter,
}


video_types = {
    '.mov': ffmpeg_video_exporter,
    '.avi': ffmpeg_video_exporter,
    '.mpg': ffmpeg_video_exporter,
    '.mpeg': ffmpeg_video_exporter,
    '.mp4': ffmpeg_video_exporter,
    '.mkv': ffmpeg_video_exporter,
    '.wmv': ffmpeg_video_exporter,
    '.gif': partial(ffmpeg_video_exporter, codec=None)
}



from .base import (
    import_image, import_images, image_paths,
    import_video, import_videos, video_paths,
    import_landmark_file, import_landmark_files, landmark_file_paths,
    import_pickle, import_pickles, pickle_paths,
    import_builtin_asset,
    menpo_data_path_to as data_path_to,
    menpo_data_dir_path as data_dir_path,
    menpo_ls_builtin_assets as ls_builtin_assets,
    register_image_importer, register_landmark_importer,
    register_pickle_importer, register_video_importer,
    same_name, same_name_video, resolve_from_paths
)



import warnings

import numpy as np
import sys
from numpy.testing import assert_allclose
import os
from pathlib import PosixPath, WindowsPath, Path
from mock import patch, PropertyMock, MagicMock
from pytest import raises


import menpo.io as mio
from menpo.io.utils import _norm_path
from menpo.image import Image
from menpo.io.output.pickle import pickle_paths_as_pure


builtins_str = '__builtin__' if sys.version_info[0] == 2 else 'builtins'

test_lg = mio.import_landmark_file(mio.data_path_to('lenna.ljson'),
                                   group='LJSON')
nan_lg = test_lg.copy()
nan_lg.points[0, :] = np.nan
test_img = Image(np.random.random([100, 100]))
colour_test_img = Image(np.random.random([3, 100, 100]))
fake_path = '/tmp/test.fake'


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('menpo.io.output.base.Path.open')
def test_export_filepath_overwrite_exists(mock_open, exists, landmark_types):
    exists.return_value = True
    landmark_types.__contains__.return_value = True
    mio.export_landmark_file(test_lg, fake_path, overwrite=True)
    mock_open.assert_called_with('wb')
    landmark_types.__getitem__.assert_called_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    assert export_function.call_count == 1


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('menpo.io.output.base.Path.open')
def test_export_filepath_no_overwrite(mock_open, exists, landmark_types):
    exists.return_value = False
    landmark_types.__contains__.return_value = True
    mio.export_landmark_file(test_lg, fake_path)
    mock_open.assert_called_with('wb')
    landmark_types.__getitem__.assert_called_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    assert export_function.call_count == 1


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('menpo.io.output.base.Path.open')
def test_export_filepath_wrong_extension(mock_open, exists, landmark_types):
    exists.return_value = False
    with raises(ValueError):
        mio.export_landmark_file(test_lg, fake_path, extension='pts')


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('menpo.io.output.base.Path.open')
def test_export_filepath_explicit_ext_no_dot(mock_open, exists, landmark_types):
    exists.return_value = False
    landmark_types.__contains__.return_value = True
    mio.export_landmark_file(test_lg, fake_path, extension='fake')
    mock_open.assert_called_with('wb')
    landmark_types.__getitem__.assert_called_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    assert export_function.call_count == 1


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('menpo.io.output.base.Path.open')
def test_export_filepath_explicit_ext_dot(mock_open, exists, landmark_types):
    exists.return_value = False
    landmark_types.__contains__.return_value = True
    mio.export_landmark_file(test_lg, fake_path, extension='.fake')
    mock_open.assert_called_with('wb')
    landmark_types.__getitem__.assert_called_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    assert export_function.call_count == 1


@patch('menpo.io.output.base.Path.exists')
def test_export_filepath_no_overwrite_exists(exists):
    exists.return_value = True
    with raises(ValueError):
        mio.export_landmark_file(test_lg, fake_path)


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
def test_export_unsupported_extension(exists, landmark_types):
    exists.return_value = False
    landmark_types.get.side_effect = KeyError
    with raises(ValueError):
        mio.export_landmark_file(test_lg, fake_path)


@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_extension_None(mock_open):
    with open(fake_path) as f:
        with raises(ValueError):
            mio.export_landmark_file(test_lg, f)


@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_extension_not_match_no_dot(mock_open, exists):
    exists.return_value = False
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        with raises(ValueError):
            mio.export_landmark_file(test_lg, f, extension='pts')


@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_extension_not_match_dot(mock_open, exists):
    exists.return_value = False
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        with raises(ValueError):
            mio.export_landmark_file(test_lg, f, extension='.pts')


@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_exists(mock_open, exists):
    exists.return_value = True
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        with raises(ValueError):
            mio.export_landmark_file(test_lg, f, extension='fake')


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_exists_overwrite(mock_open, exists,
                                                  landmark_types):
    exists.return_value = True
    landmark_types.__contains__.return_value = True
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test_lg, f, overwrite=True, extension='fake')
    landmark_types.__getitem__.assert_called_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    assert export_function.call_count == 1


@patch('menpo.io.output.base.landmark_types')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_file_handle_file_non_file_buffer(mock_open, exists,
                                                 landmark_types):
    exists.return_value = False
    landmark_types.__contains__.return_value = True
    with open(fake_path) as f:
        del f.name  # Equivalent to raising an AttributeError side effect
        mio.export_landmark_file(test_lg, f, extension='fake')
    landmark_types.__getitem__.assert_called_with('.fake')
    export_function = landmark_types.__getitem__.return_value
    assert export_function.call_count == 1


@patch('menpo.io.output.landmark.json.dump')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_landmark_ljson(mock_open, exists, json_dump):
    exists.return_value = False
    fake_path = '/fake/fake.ljson'
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test_lg, f, extension='ljson')
    assert json_dump.call_count == 1


@patch('menpo.io.output.landmark.json.dump')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_landmark_ljson_3d(mock_open, exists, json_dump):
    exists.return_value = False
    fake_path = '/fake/fake3d.ljson'
    test3d_lg = test_lg.copy()
    fake_z_points = np.random.random(test3d_lg.points.shape[0])
    test3d_lg.points = np.concatenate([
        test3d_lg.points, fake_z_points[..., None]], axis=-1)

    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test3d_lg, f, extension='ljson')

    assert json_dump.call_count == 1
    json_points = np.array(json_dump.call_args[0][0]['groups']['LJSON']['landmarks']['points'])
    assert_allclose(json_points[:, -1], fake_z_points)


@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_landmark_ljson_nan_values(mock_open, exists):
    exists.return_value = False
    mock_writer = MagicMock()
    mock_open.return_value.__enter__.return_value = mock_writer
    fake_path = '/fake/fake.ljson'
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(nan_lg, f, extension='ljson')

    # yeah this is grim, but it should work.
    assert 'null' in '{}'.format(mock_open.mock_calls)


@patch('menpo.io.output.landmark.np.savetxt')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_landmark_pts(mock_open, exists, save_txt):
    exists.return_value = False
    fake_path = '/fake/fake.pts'
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test_lg, f, extension='pts')
    assert save_txt.call_count == 1


@patch('menpo.image.base.PILImage')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_image_jpg(mock_open, exists, PILImage):
    exists.return_value = False
    fake_path = '/fake/fake.jpg'
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_image(test_img, f, extension='jpg')
    assert PILImage.fromarray.return_value.save.call_count == 1


@patch('subprocess.Popen')
@patch('menpo.io.output.base.Path.exists')
def test_export_video_avi_gray(exists, pipe):
    exists.return_value = False
    fake_path = Path('/fake/fake.avi')
    mio.export_video([test_img, test_img], fake_path, extension='avi')
    assert pipe.return_value.stdin.write.call_count == 2
    assert 'gray8' in pipe.call_args[0][0]


@patch('subprocess.Popen')
@patch('menpo.io.output.base.Path.exists')
def test_export_video_avi_colour(exists, pipe):
    exists.return_value = False
    fake_path = Path('/fake/fake.avi')
    mio.export_video([colour_test_img, colour_test_img], fake_path,
                     extension='avi')
    assert pipe.return_value.stdin.write.call_count == 2
    assert 'rgb24' in pipe.call_args[0][0]


@patch('subprocess.Popen')
@patch('menpo.io.output.base.Path.exists')
def test_export_video_avi_gray_first_mixed(exists, pipe):
    exists.return_value = False
    fake_path = Path('/fake/fake.avi')
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        mio.export_video([test_img, colour_test_img], fake_path,
                         extension='avi')
        assert len(w) == 1
    assert pipe.return_value.stdin.write.call_count == 2
    assert 'gray8' in pipe.call_args[0][0]


@patch('subprocess.Popen')
@patch('menpo.io.output.base.Path.exists')
def test_export_video_avi_colour_first_mixed(exists, pipe):
    exists.return_value = False
    fake_path = Path('/fake/fake.avi')
    mio.export_video([colour_test_img, test_img], fake_path,
                     extension='avi')
    assert pipe.return_value.stdin.write.call_count == 2
    recon_0 = pipe.return_value.stdin.write.mock_calls[0][1][0]
    assert np.frombuffer(recon_0, dtype=np.uint8).size == 30000
    # Ensure that the second frame is converted to gray
    recon_1 = pipe.return_value.stdin.write.mock_calls[1][1][0]
    assert np.frombuffer(recon_1, dtype=np.uint8).size == 30000
    assert 'rgb24' in pipe.call_args[0][0]


@patch('subprocess.Popen')
@patch('menpo.io.output.base.Path.exists')
def test_export_video_avi_shape_mismatch(exists, pipe):
    exists.return_value = False
    fake_path = Path('/fake/fake.avi')
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        mio.export_video([test_img.resize([150, 100]), test_img], fake_path,
                         extension='avi')
        assert len(w) == 1
    assert 'gray8' in pipe.call_args[0][0]


@patch('subprocess.Popen')
@patch('menpo.io.output.base.Path.exists')
def test_export_video_gif(exists, pipe):
    exists.return_value = False
    fake_path = Path('/fake/fake.gif')
    mio.export_video([test_img, test_img], fake_path, extension='gif')
    assert pipe.return_value.stdin.write.call_count == 2


@patch('subprocess.Popen')
@patch('menpo.io.output.base.Path.exists')
def test_export_video_avi_kwargs(exists, pipe):
    exists.return_value = False
    fake_path = Path('/fake/fake.avi')
    mio.export_video([test_img, test_img], fake_path, extension='avi', **{'crf' : '0'})
    assert pipe.return_value.stdin.write.call_count == 2
    assert '-crf' in pipe.call_args[0][0]


@patch('menpo.io.output.pickle.pickle.dump')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(__name__), create=True)
def test_export_pickle(mock_open, exists, pickle_dump):
    exists.return_value = False
    fake_path = '/fake/fake.pkl'
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_pickle(test_lg, f)
    assert pickle_dump.call_count == 1


@patch('menpo.io.output.pickle.pickle.dump')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(builtins_str))
def test_export_pickle_with_path_uses_open(mock_open, exists, pickle_dump):
    exists.return_value = False
    fake_path = str(_norm_path('fake.pkl.gz'))
    mock_open_enter = MagicMock()
    # Make sure the name attribute returns the path
    mock_open_enter.__enter__.return_value.configure_mock(name=fake_path)
    mock_open.return_value = mock_open_enter
    mio.export_pickle(test_lg, fake_path)
    assert pickle_dump.call_count == 1
    mock_open.assert_called_with(fake_path, 'wb')


@patch('menpo.io.output.pickle.pickle.dump')
@patch('menpo.io.output.base.Path.exists')
@patch('{}.open'.format(builtins_str))
def test_export_pickle_with_path_expands_vars(mock_open, exists, pickle_dump):
    exists.return_value = False
    fake_path = '~/fake/fake.pkl.gz'
    mock_open_enter = MagicMock()
    # Make sure the name attribute returns the path
    mock_open_enter.__enter__.return_value.configure_mock(name=fake_path)
    mock_open.return_value = mock_open_enter
    mio.export_pickle(test_lg, fake_path)
    assert pickle_dump.call_count == 1
    expected_path = os.path.join(os.path.expanduser('~'), 'fake', 'fake.pkl.gz')
    mock_open.assert_called_with(expected_path, 'wb')


def test_pickle_paths_as_pure_switches_reduce_method_on_path():
    prev_reduce = Path.__reduce__
    with pickle_paths_as_pure():
        assert prev_reduce != Path.__reduce__
    assert prev_reduce == Path.__reduce__


def test_pickle_paths_as_pure_switches_reduce_method_on_posix_path():
    prev_reduce = PosixPath.__reduce__
    with pickle_paths_as_pure():
        assert prev_reduce != PosixPath.__reduce__
    assert prev_reduce == PosixPath.__reduce__


def test_pickle_paths_as_pure_switches_reduce_method_on_windows_path():
    prev_reduce = WindowsPath.__reduce__
    with pickle_paths_as_pure():
        assert prev_reduce != WindowsPath.__reduce__
    assert prev_reduce == WindowsPath.__reduce__


def test_pickle_paths_as_pure_cleans_up_on_exception():
    prev_reduce = Path.__reduce__
    try:
        with pickle_paths_as_pure():
            raise ValueError()
    except ValueError:
        assert prev_reduce == Path.__reduce__  # ensure we clean up



from __future__ import division
import numpy as np
from scipy.sparse import issparse
from .linalg import dot_inplace_right


def eigenvalue_decomposition(C, is_inverse=False, eps=1e-10):
    r"""
    Eigenvalue decomposition of a given covariance (or scatter) matrix.

    Parameters
    ----------
    C : ``(N, N)`` `ndarray` or `scipy.sparse`
        The Covariance/Scatter matrix. If it is a `numpy.array`, then
        `numpy.linalg.eigh` is used. If it is an instance of `scipy.sparse`,
        then `scipy.sparse.linalg.eigsh` is used. If it is a precision matrix
        (inverse covariance), then set `is_inverse=True`.
    is_inverse : `bool`, optional
        It ``True``, then it is assumed that `C` is a precision matrix (
        inverse covariance). Thus, the eigenvalues will be inverted. If
        ``False``, then it is assumed that `C` is a covariance matrix.
    eps : `float`, optional
        Tolerance value for positive eigenvalue. Those eigenvalues smaller
        than the specified eps value, together with their corresponding
        eigenvectors, will be automatically discarded. The final
        limit is computed as ::

            limit = np.max(np.abs(eigenvalues)) * eps

    Returns
    -------
    pos_eigenvectors : ``(N, p)`` `ndarray`
        The matrix with the eigenvectors corresponding to positive eigenvalues.
    pos_eigenvalues : ``(p,)`` `ndarray`
        The array of positive eigenvalues.
    """
    # compute eigenvalue decomposition
    if issparse(C):
        from scipy.sparse.linalg import eigsh
        eigenvalues, eigenvectors = eigsh(C, k=C.shape[0] - 1)
    else:
        eigenvalues, eigenvectors = np.linalg.eigh(C)

    # sort eigenvalues from largest to smallest
    index = np.argsort(eigenvalues)[::-1]
    eigenvalues = eigenvalues[index]
    eigenvectors = eigenvectors[:, index]

    # set tolerance limit
    limit = np.max(np.abs(eigenvalues)) * eps

    # select positive eigenvalues
    pos_index = eigenvalues > 0.0
    pos_eigenvalues = eigenvalues[pos_index]
    pos_eigenvectors = eigenvectors[:, pos_index]
    # check they are within the expected tolerance
    index = pos_eigenvalues > limit
    pos_eigenvalues = pos_eigenvalues[index]
    pos_eigenvectors = pos_eigenvectors[:, index]

    # if C was a precision matrix (inverse covariance), then invert and re-sort
    # the eigenvalues
    if is_inverse:
        pos_eigenvalues = pos_eigenvalues[::-1] ** -1
        pos_eigenvectors = pos_eigenvectors[:, ::-1]

    return pos_eigenvectors, pos_eigenvalues


def pca(X, centre=True, inplace=False, eps=1e-10):
    r"""
    Apply Principal Component Analysis (PCA) on the data matrix `X`. In the case
    where the data matrix is very large, it is advisable to set
    ``inplace = True``. However, note this destructively edits the data matrix
    by subtracting the mean inplace.

    Parameters
    ----------
    X : ``(n_samples, n_dims)`` `ndarray`
        Data matrix.
    centre : `bool`, optional
        Whether to centre the data matrix. If `False`, zero will be subtracted.
    inplace : `bool`, optional
        Whether to do the mean subtracting inplace or not. This is crucial if
        the data matrix is greater than half the available memory size.
    eps : `float`, optional
        Tolerance value for positive eigenvalue. Those eigenvalues smaller
        than the specified eps value, together with their corresponding
        eigenvectors, will be automatically discarded.

    Returns
    -------
    U (eigenvectors) : ``(``(n_components, n_dims)``)`` `ndarray`
        Eigenvectors of the data matrix.
    l (eigenvalues) : ``(n_components,)`` `ndarray`
        Positive eigenvalues of the data matrix.
    m (mean vector) : ``(n_dimensions,)`` `ndarray`
        Mean that was subtracted from the data matrix.
    """
    n, d = X.shape

    if centre:
        # centre data
        # m (mean vector): d
        m = np.mean(X, axis=0)
    else:
        m = np.zeros(d, dtype=X.dtype)

    # This is required if the data matrix is very large!
    if inplace:
        X -= m
    else:
        X = X - m

    if d < n:
        # compute covariance matrix
        # C (covariance): d x d
        C = np.dot(X.conj().T, X) / (n - 1)
        # C should be perfectly symmetrical, but numerical error can creep
        # in. Enforce symmetry here to avoid creating complex eigenvectors
        C = (C + C.conj().T) / 2.0

        # perform eigenvalue decomposition
        # U (eigenvectors): d x n
        # s (eigenvalues):  n
        U, l = eigenvalue_decomposition(C, is_inverse=False, eps=eps)

        # transpose U
        # U: n x d
        U = U.T

    else:
        # d > n
        # compute small covariance matrix
        # C (covariance): n x n
        C = np.dot(X, X.conj().T) / (n - 1)
        # C should be perfectly symmetrical, but numerical error can creep
        # in. Enforce symmetry here to avoid creating complex eigenvectors
        C = (C + C.conj().T) / 2.0

        # perform eigenvalue decomposition
        # V (eigenvectors): n x n
        # s (eigenvalues):  n
        V, l = eigenvalue_decomposition(C, is_inverse=False, eps=eps)

        # compute final eigenvectors
        # U: n x d
        w = np.sqrt(1.0 / ((n - 1) * l))
        dot = dot_inplace_right if inplace else np.dot
        U = dot(V.conj().T, X)
        U *= w[:, None]

    return U, l, m


# The default value of eps tolerance is set to 1e-5 (instead of 1e-10 that used
# to be). This is done in order for pcacov to work for inverse single precision C
# i.e. is_inverse=True and dtype=np.float32. 1e-10 works perfectly when the
# covariance matrix has double precision (np.float64). However, if C has single
# precision (np.float32) and is inverse, then the first two eigenvectors end up
# having noise.
def pcacov(C, is_inverse=False, eps=1e-5):
    r"""
    Apply Principal Component Analysis (PCA) given a covariance/scatter matrix
    `C`. In the case where the data matrix is very large, it is advisable to set
    ``inplace = True``. However, note this destructively edits the data matrix
    by subtracting the mean inplace.

    Parameters
    ----------
    C : ``(N, N)`` `ndarray` or `scipy.sparse`
        The Covariance/Scatter matrix. If it is a precision matrix (inverse
        covariance), then set `is_inverse=True`.
    is_inverse : `bool`, optional
        It ``True``, then it is assumed that `C` is a precision matrix (
        inverse covariance). Thus, the eigenvalues will be inverted. If
        ``False``, then it is assumed that `C` is a covariance matrix.
    eps : `float`, optional
        Tolerance value for positive eigenvalue. Those eigenvalues smaller
        than the specified eps value, together with their corresponding
        eigenvectors, will be automatically discarded.

    Returns
    -------
    U (eigenvectors) : ``(n_components, n_dims)`` `ndarray`
        Eigenvectors of the data matrix.
    l (eigenvalues) : ``(n_components,)`` `ndarray`
        Positive eigenvalues of the data matrix.
    """
    if C.shape[0] != C.shape[1]:
        raise ValueError('C must be square.')

    # C should be perfectly symmetrical, but numerical error can creep in.
    # Enforce symmetry here to avoid creating complex eigenvectors
    C = (C + C.conj().T) / 2.0

    # C (covariance): d x d
    # perform eigenvalue decomposition
    # U (eigenvectors): d x n
    # s (eigenvalues):  n
    U, l = eigenvalue_decomposition(C, is_inverse=is_inverse, eps=eps)

    # transpose U
    # U: n x d
    U = U.conj().T

    return U, l


def ipca(B, U_a, l_a, n_a, m_a=None, f=1.0, eps=1e-10):
    r"""
    Perform Incremental PCA on the eigenvectors ``U_a``, eigenvalues ``l_a`` and
    mean vector ``m_a`` (if present) given a new data matrix ``B``.

    Parameters
    ----------
    B : ``(n_samples, n_dims)`` `ndarray`
        New data matrix.
    U_a : ``(n_components, n_dims)`` `ndarray`
        Eigenvectors to be updated.
    l_a : (n_components) `ndarray`
        Eigenvalues to be updated.
    n_a : `int`
        Total number of samples used to produce U_a, s_a and m_a.
    m_a : ``(n_dims,)`` `ndarray`, optional
        Mean to be updated. If ``None`` or ``(n_dims,)`` `ndarray` filled
        with 0s the data matrix will not be centred.
    f : ``[0, 1]`` `float`, optional
        Forgetting factor that weights the relative contribution of new
        samples vs old samples. If 1.0, all samples are weighted equally
        and, hence, the results is the exact same as performing batch
        PCA on the concatenated list of old and new simples. If <1.0,
        more emphasis is put on the new samples. See [1] for details.
    eps : `float`, optional
        Tolerance value for positive eigenvalue. Those eigenvalues smaller
        than the specified eps value, together with their corresponding
        eigenvectors, will be automatically discarded.

    Returns
    -------
    U (eigenvectors) : ``(n_components, n_dims)`` `ndarray`
        Updated eigenvectors.
    s (eigenvalues) : ``(n_components,)`` `ndarray`
        Updated positive eigenvalues.
    m (mean vector) : ``(n_dims,)`` `ndarray`
        Updated mean.

    References
    ----------
    .. [1] David Ross, Jongwoo Lim, Ruei-Sung Lin, Ming-Hsuan Yang.
       "Incremental Learning for Robust Visual Tracking". IJCV, 2007.
    """
    # multiply current eigenvalues by total number of samples and square
    # root them to obtain singular values of the original data.
    s_a = np.sqrt((n_a - 1) * l_a)

    # obtain number of dimensions and number of samples of new data.
    n_b, d = B.shape
    # multiply the number of samples of the original data by the forgetting
    # factor
    n_a *= f
    # total number of samples
    n = n_a + n_b

    if m_a is not None and not np.all(m_a == 0):
        # centred ipca; compute mean of new data
        m_b = np.mean(B, axis=0)
        # compute new mean
        m = (n_a / n) * m_a + (n_b / n) * m_b
        # centre new data
        B = B - m_b
        # augment centred data with extra sample
        B = np.vstack((B, np.sqrt((n_a * n_b) / n) * (m_b - m_a)))
    else:
        m = np.zeros(d, dtype=B.dtype)

    # project out current eigenspace out of data matrix
    PB = B - B.dot(U_a.T).dot(U_a)
    # orthogonalise the previous projection using QR
    B_tilde = np.linalg.qr(PB.T)[0].T

    # form R matrix
    S_a = np.diag(s_a)
    R = np.hstack((np.vstack((f * S_a, B.dot(U_a.T))),
                   np.vstack((np.zeros((S_a.shape[0], B_tilde.shape[0]),
                                       dtype=B.dtype),
                              PB.dot(B_tilde.T)))))

    # compute SVD of R
    U_tilde, s_tilde, Vt_tilde = np.linalg.svd(R)

    # compute new eigenvalues
    l = s_tilde ** 2 / (n - 1)
    # keep only positive eigenvalues within tolerance
    l = l[l > eps]

    U = Vt_tilde.dot(np.vstack((U_a, B_tilde)))[:len(l), :]

    return U, l, m



from __future__ import division
import numpy as np
from menpo.base import MenpoMissingDependencyError, partial_doc
from menpo.image.base import normalize_pixels_range
from ..base import winitfeature

try:
    from cyvlfeat.sift.dsift import dsift as cyvlfeat_dsift
except ImportError as e:
    raise MenpoMissingDependencyError(e)


@winitfeature
def dsift(pixels, window_step_horizontal=1, window_step_vertical=1,
          num_bins_horizontal=2, num_bins_vertical=2, num_or_bins=9,
          cell_size_horizontal=6, cell_size_vertical=6, fast=True,
          verbose=False):
    r"""
    Computes a 2-dimensional dense SIFT features image with ``C`` number of
    channels, where
    ``C = num_bins_horizontal * num_bins_vertical * num_or_bins``. The dense
    SIFT [2]_ implementation is taken from Vlfeat [1]_.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, Y, X)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels.
    window_step_horizontal : `int`, optional
        Defines the horizontal step by which the window is moved, thus it
        controls the features density. The metric unit is pixels.
    window_step_vertical : `int`, optional
        Defines the vertical step by which the window is moved, thus it
        controls the features density. The metric unit is pixels.
    num_bins_horizontal : `int`, optional
        Defines the number of histogram bins in the X direction.
    num_bins_vertical : `int`, optional
        Defines the number of histogram bins in the Y direction.
    num_or_bins : `int`, optional
        Defines the number of orientation histogram bins.
    cell_size_horizontal : `int`, optional
        Defines cell width in pixels. The cell is the region that is covered by
        a spatial bin.
    cell_size_vertical : `int`, optional
        Defines cell height in pixels. The cell is the region that is covered by
        a spatial bin.
    fast : `bool`, optional
        If ``True``, then the windowing function is a piecewise-flat, rather
        than Gaussian. While this breaks exact SIFT equivalence, in practice it
        is much faster to compute.
    verbose : `bool`, optional
        Flag to print SIFT related information.

    Raises
    ------
    ValueError
        Only 2D arrays are supported
    ValueError
        Size must only contain positive integers.
    ValueError
        Step must only contain positive integers.
    ValueError
        Window size must be a positive integer.
    ValueError
        Geometry must only contain positive integers.

    References
    ----------
    .. [1] Vedaldi, Andrea, and Brian Fulkerson. "VLFeat: An open and portable
       library of computer vision algorithms." Proceedings of the international
       conference on Multimedia. ACM, 2010.
    .. [2] Lowe, David G. "Distinctive image features from scale-invariant
       keypoints." International journal of computer vision 60.2 (2004): 91-110.
    """
    # If norm is set to True, then the centers array will have a third column
    # with descriptor norm, or energy, before contrast normalization.
    # This information can be used to suppress low contrast descriptors.
    centers, output = cyvlfeat_dsift(
        pixels[0], step=[window_step_vertical, window_step_horizontal],
        size=[cell_size_vertical, cell_size_horizontal], bounds=None,
        norm=False, fast=fast, float_descriptors=True,
        geometry=(num_bins_vertical, num_bins_horizontal, num_or_bins),
        verbose=False)

    # the output shape can be calculated from looking at the range of
    # centres / the window step size in each dimension. Note that cyvlfeat
    # returns x, y centres.
    shape = (((centers[-1, :] - centers[0, :]) /
              [window_step_vertical, window_step_horizontal]) + 1)
    shape = shape.astype(np.int)

    # print information
    if verbose:
        info_str = "Dense SIFT features:\n" \
                   "  - Input image is {}W x {}H with {} channels.\n" \
                   "  - Sampling step of ({}W, {}H).\n" \
                   "  - {}W x {}H spatial bins and {} orientation bins.\n" \
                   "  - Cell size of {}W x {}H pixels.\n".format(
                   pixels.shape[2], pixels.shape[1], pixels.shape[0],
                   window_step_horizontal, window_step_vertical,
                   num_bins_horizontal, num_bins_vertical, num_or_bins,
                   cell_size_horizontal, cell_size_vertical)
        if fast:
            info_str += "  - Fast mode is enabled.\n"
        info_str += "Output image size {}W x {}H x {}.".format(
            int(shape[1]), int(shape[0]), output.shape[0])
        print(info_str)

    # return SIFT and centers in the correct form
    return (np.require(np.rollaxis(output.reshape((shape[0], shape[1], -1)),
                                   -1),
                       dtype=np.double, requirements=['C']),
            np.require(centers.reshape((shape[0], shape[1], -1)),
                       dtype=np.int))


# A predefined method for a 'faster' dsift method
fast_dsift = partial_doc(dsift, fast=True, cell_size_vertical=5,
                         cell_size_horizontal=5, num_bins_horizontal=1,
                         num_bins_vertical=1, num_or_bins=8)


# Predefined dsift that returns a 128d vector
def vector_128_dsift(x, dtype=np.float32):
    r"""
    Computes a SIFT feature vector from a square patch (or image). Patch
    **must** be square and the output vector will *always* be a ``(128,)``
    vector. Please see :func:`dsift` for more information.

    Parameters
    ----------
    x : :map:`Image` or subclass or ``(C, Y, Y)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels. Must be square i.e.
        ``height == width``.
    dtype : ``np.dtype``, optional
        The dtype of the returned vector.

    Raises
    ------
    ValueError
        Only square images are supported.
    """
    if not isinstance(x, np.ndarray):
        x = x.pixels
    if x.shape[-1] != x.shape[-2]:
        raise ValueError('This feature only works with square images '
                         'i.e. width == height')
    patch_shape = x.shape[-1]
    n_bins = 4
    c_size = patch_shape // n_bins
    x = normalize_pixels_range(x, error_on_unknown_type=False)
    return dsift(x,
                 window_step_horizontal=patch_shape,
                 window_step_vertical=patch_shape,
                 num_bins_horizontal=n_bins, num_bins_vertical=n_bins,
                 cell_size_horizontal=c_size, cell_size_vertical=c_size,
                 num_or_bins=8, fast=True).astype(dtype)


# Predefined dsift that returns a 128d vector normalized by the hellinger norm
def hellinger_vector_128_dsift(x, dtype=np.float32):
    r"""
    Computes a SIFT feature vector from a square patch (or image). Patch
    **must** be square and the output vector will *always* be a ``(128,)``
    vector. Please see :func:`dsift` for more information.

    The output of :func:`vector_128_dsift` is normalized using the hellinger
    norm (also called the Bhattacharyya distance) which is a measure
    designed to quantify the similarity between two probability distributions.
    Since SIFT is a histogram based feature, this has been shown to improve
    performance. Please see [1]_ for more information.

    Parameters
    ----------
    x : :map:`Image` or subclass or ``(C, Y, Y)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels. Must be square i.e.
        ``height == width``.
    dtype : ``np.dtype``, optional
        The dtype of the returned vector.

    Raises
    ------
    ValueError
        Only square images are supported.

    References
    ----------
    .. [1] Arandjelovic, Relja, and Andrew Zisserman. "Three things everyone
           should know to improve object retrieval.", CVPR, 2012.
    """
    h = vector_128_dsift(x, dtype=dtype)
    h /= (h.sum(axis=0) + 1e-15)
    return np.sqrt(h)



from . import base

from . import feature
from . import image
from . import io
from . import landmark
from . import math
from . import model
from . import shape
from . import transform
from . import visualize

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import numpy as np
from pytest import raises
from numpy.testing import assert_allclose

from menpo.transform import Affine, NonUniformScale


def test_affine_identity_2d():
    assert_allclose(Affine.init_identity(2).h_matrix, np.eye(3))


def test_affine_identity_3d():
    assert_allclose(Affine.init_identity(3).h_matrix, np.eye(4))


def test_basic_2d_affine():
    linear_component = np.array([[1, -6],
                                 [-3, 2]])
    translation_component = np.array([7, -8])
    h_matrix = np.eye(3, 3)
    h_matrix[:-1, :-1] = linear_component
    h_matrix[:-1, -1] = translation_component
    affine = Affine(h_matrix)
    x = np.array([[0, 1],
                  [1, 1],
                  [-1, -5],
                  [3, -5]])
    # transform x explicitly
    solution = np.dot(x, linear_component.T) + translation_component
    # transform x using the affine transform
    result = affine.apply(x)
    # check that both answers are equivalent
    assert_allclose(solution, result)
    # create several copies of x
    x_copies = np.array([x, x, x, x, x, x, x, x])
    # transform all of copies at once using the affine transform
    results = affine.apply(x_copies)
    # check that all copies have been transformed correctly
    for r in results:
        assert_allclose(solution, r)


def test_basic_3d_affine():
    linear_component = np.array([[1, 6, -4],
                                 [-3, -2, 5],
                                 [5, -1, 3]])
    translation_component = np.array([7, -8, 9])
    h_matrix = np.eye(4, 4)
    h_matrix[:-1, :-1] = linear_component
    h_matrix[:-1, -1] = translation_component
    affine = Affine(h_matrix)
    x = np.array([[0, 1, 2],
                  [1, 1, 1],
                  [-1, 2, -5],
                  [1, -5, -1]])
    # transform x explicitly
    solution = np.dot(x, linear_component.T) + translation_component
    # transform x using the affine transform
    result = affine.apply(x)
    # check that both answers are equivalent
    assert_allclose(solution, result)
    # create several copies of x
    x_copies = np.array([x, x, x, x, x, x, x, x])
    # transform all of copies at once using the affine transform
    results = affine.apply(x_copies)
    # check that all copies have been transformed correctly
    for r in results:
        assert_allclose(solution, r)


def test_affine_2d_n_parameters():
    homo = np.eye(3)
    t = Affine(homo)
    assert (t.n_parameters == 6)


def test_affine_3d_n_parameters():
    homo = np.eye(4)
    t = Affine(homo)
    assert (t.n_parameters == 12)


def test_affine_non_square_h_matrix():
    homo = np.random.rand(4, 6)
    with raises(ValueError):
        Affine(homo)


def test_affine_incorrect_bottom_row():
    homo = np.random.rand(4, 4)
    with raises(ValueError):
        Affine(homo)


def test_affine_non_square_h_matrix():
    homo = np.random.rand(4, 6)
    with raises(ValueError):
        Affine(homo)


def test_affine_compose_inplace_affine():
    a = Affine.init_identity(2)
    b = Affine.init_identity(2)
    a.compose_before_inplace(b)
    assert(np.all(a.h_matrix == b.h_matrix))


def test_affine_pseudoinverse():
    s = NonUniformScale([4, 3])
    inv_man = NonUniformScale([1./4, 1./3])
    b = Affine(s.h_matrix)
    i = b.pseudoinverse()
    assert_allclose(i.h_matrix, inv_man.h_matrix)



from collections import OrderedDict
import numpy as np
from pytest import raises
from numpy.testing import assert_allclose
from scipy.sparse import csr_matrix

from menpo.shape import LabelledPointUndirectedGraph, PointUndirectedGraph
from menpo.testing import is_same_array


points = np.ones((10, 3))
adjacency_matrix = csr_matrix((10, 10))
mask_dict = OrderedDict([('all', np.ones(10, dtype=np.bool))])
lower = np.array([True, True, True, True, True, True,
                  False, False, False, False], dtype=np.bool)
upper = np.array([False, False, False, False, False, False,
                  True, True, True, True], dtype=np.bool)
mask_dict_2 = OrderedDict([('lower', lower), ('upper', upper)])
mask_dict_3 = OrderedDict([('all', np.ones(10, dtype=np.bool)),
                           ('lower', lower), ('upper', upper)])


def test_LabelledPointUndirectedGraph_copy_true():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict)
    assert not is_same_array(lgroup.points, points)
    assert lgroup._labels_to_masks is not mask_dict
    assert lgroup.adjacency_matrix is not adjacency_matrix


def test_LabelledPointUndirectedGraph_copy_false():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict, copy=False)
    assert is_same_array(lgroup.points, points)
    assert lgroup._labels_to_masks is mask_dict
    assert lgroup.adjacency_matrix is adjacency_matrix


def test_LabelledPointUndirectedGraph_copy_method():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict)
    lgroup_copy = lgroup.copy()

    assert not is_same_array(lgroup_copy.points, lgroup.points)
    # Check the mask dictionary is deepcopied properly
    assert lgroup._labels_to_masks is not lgroup_copy._labels_to_masks
    masks = zip(lgroup_copy._labels_to_masks.values(),
                lgroup._labels_to_masks.values())
    for ms in masks:
        assert ms[0] is not ms[1]


def test_LabelledPointUndirectedGraph_iterate_labels():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict)

    for l in lgroup.labels:
        assert l == 'all'
    assert len(lgroup.labels) == 1


def test_LabelledPointUndirectedGraph_get_label():
    pcloud = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict_2)

    assert isinstance(pcloud, PointUndirectedGraph)
    assert pcloud.get_label('lower').n_points == 6
    assert pcloud.get_label('upper').n_points == 4


def test_LabelledPointUndirectedGraph_add_label():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict_2)

    new_lgroup = lgroup.add_label('lower2', [0, 1])
    assert not is_same_array(new_lgroup.points, lgroup.points)

    lower_pcloud = new_lgroup.get_label('lower2')
    assert lower_pcloud.n_points == 2
    assert_allclose(lower_pcloud.points[0, :], [1., 1., 1.])
    assert_allclose(lower_pcloud.points[1, :], [1., 1., 1.])


def test_LabelledPointUndirectedGraph_add_ordered_labels():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict_2)

    labels = lgroup.labels
    assert labels[0] == 'lower'
    assert labels[1] == 'upper'

    new_lgroup = lgroup.add_label('A', [0, 1])
    new_labels = new_lgroup.labels

    assert new_labels[2] == 'A'


def test_LabelledPointUndirectedGraph_remove_label():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict_3)

    new_lgroup = lgroup.remove_label('lower')

    assert 'all' in new_lgroup.labels
    assert 'lower' not in new_lgroup.labels
    assert 'all' in lgroup.labels


def test_LabelledPointUndirectedGraph_remove_label_unlabelled():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict_2)

    with raises(ValueError):
        lgroup.remove_label('lower')


def test_LabelledPointUndirectedGraph_create_unlabelled():
    points = np.array([[0, 1], [2, 3], [4, 5]])
    adjacency_matrix = csr_matrix((3, 3))
    mask_dict = OrderedDict([('all', np.zeros(3, dtype=np.bool))])

    with raises(ValueError):
        LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict)


def test_LabelledPointUndirectedGraph_pass_non_ordered_dict():
    points = np.array([[0, 1], [2, 3], [4, 5]])
    adjacency_matrix = csr_matrix((3, 3))
    mask_dict = {'all': np.ones(3, dtype=np.bool)}

    with raises(ValueError):
        LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict)


def test_LabelledPointUndirectedGraph_create_no_mask():
    points = np.array([[0, 1], [2, 3], [4, 5]])
    adjacency_matrix = csr_matrix((3, 3))
    mask_dict = None

    with raises(ValueError):
        LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict)


def test_LabelledPointUndirectedGraph_create_incorrect_shape():
    points = np.array([[0, 1], [2, 3], [4, 5]])
    adjacency_matrix = csr_matrix((3, 3))
    mask_dict = OrderedDict([('all', np.ones(5, dtype=np.bool))])

    with raises(ValueError):
        LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict)


def test_LabelledPointUndirectedGraph_with_labels():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict_2)

    new_lgroup = lgroup.with_labels('lower')

    assert new_lgroup.n_labels == 1
    assert new_lgroup.n_points == 6
    assert 'lower' in new_lgroup.labels

    new_lgroup = lgroup.with_labels(['lower'])

    assert new_lgroup.n_labels == 1
    assert new_lgroup.n_points == 6
    assert 'lower' in new_lgroup.labels


def test_LabelledPointUndirectedGraph_without_labels():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict_2)

    new_lgroup = lgroup.without_labels('upper')

    assert new_lgroup.n_labels == 1
    assert new_lgroup.n_points == 6
    assert 'lower' in new_lgroup.labels
    assert 'upper' not in new_lgroup.labels

    new_lgroup = lgroup.without_labels(['upper'])

    assert new_lgroup.n_labels == 1
    assert new_lgroup.n_points == 6
    assert 'lower' in new_lgroup.labels
    assert 'upper' not in new_lgroup.labels


def test_LabelledPointUndirectedGraph_str():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict_2)

    out_str = lgroup.__str__()
    assert len(out_str) > 0


def test_LabelledPointUndirectedGraph_create_with_all_label():
    lgroup = LabelledPointUndirectedGraph.init_with_all_label(points,
                                                              adjacency_matrix)

    assert lgroup.n_labels == 1
    assert 'all' in lgroup.labels


def test_LabelledPointUndirectedGraph_has_nan_values():
    points2 = points.copy()
    points2[0, 0] = np.nan
    lgroup = LabelledPointUndirectedGraph.init_with_all_label(points2,
                                                              adjacency_matrix)
    assert lgroup.has_nan_values()



import numpy as np

from menpo.image import Image
from menpo.shape import PointCloud, TriMesh, TexturedTriMesh, ColouredTriMesh
from menpo.testing import is_same_array


def test_trimesh_copy():
    points = np.ones([10, 3])
    trilist = np.ones([10, 3])
    landmarks = PointCloud(np.ones([3, 3]), copy=False)

    tmesh = TriMesh(points, trilist=trilist, copy=False)
    tmesh.landmarks['test'] = landmarks
    tmesh_copy = tmesh.copy()

    assert (not is_same_array(tmesh_copy.points, tmesh.points))
    assert (not is_same_array(tmesh_copy.trilist, tmesh.trilist))
    assert (not is_same_array(tmesh_copy.landmarks['test'].points,
                              tmesh.landmarks['test'].points))


def test_colouredtrimesh_copy():
    points = np.ones([10, 3])
    colours = np.ones([10, 3])
    trilist = np.ones([10, 3])
    landmarks = PointCloud(np.ones([3, 3]), copy=False)

    ctmesh = ColouredTriMesh(points, trilist=trilist, colours=colours,
                             copy=False)
    ctmesh.landmarks['test'] = landmarks
    ctmesh_copy = ctmesh.copy()

    assert (not is_same_array(ctmesh_copy.points, ctmesh.points))
    assert (not is_same_array(ctmesh_copy.trilist, ctmesh.trilist))
    assert (not is_same_array(ctmesh_copy.colours, ctmesh.colours))
    assert (not is_same_array(ctmesh_copy.landmarks['test'].points,
                              ctmesh.landmarks['test'].points))


def test_texturedtrimesh_copy():
    points = np.ones([10, 3])
    tcoords = np.ones([10, 3])
    trilist = np.ones([10, 3])
    landmarks = PointCloud(np.ones([3, 3]), copy=False)
    landmarks_im = PointCloud(np.ones([3, 2]), copy=False)
    pixels = np.ones([10, 10, 1])
    texture = Image(pixels, copy=False)
    texture.landmarks['test_im'] = landmarks_im

    ttmesh = TexturedTriMesh(points, tcoords, texture, trilist=trilist,
                             copy=False)
    ttmesh.landmarks['test'] = landmarks
    ttmesh_copy = ttmesh.copy()

    assert (not is_same_array(ttmesh_copy.points, ttmesh.points))
    assert (not is_same_array(ttmesh_copy.trilist, ttmesh.trilist))
    assert (not is_same_array(ttmesh_copy.tcoords.points,
                              ttmesh.tcoords.points))
    assert (not is_same_array(ttmesh_copy.texture.pixels,
                              ttmesh.texture.pixels))
    assert (not is_same_array(
        ttmesh_copy.texture.landmarks['test_im'].points,
        ttmesh.texture.landmarks['test_im'].points))
    assert (not is_same_array(ttmesh_copy.landmarks['test'].points,
                              ttmesh.landmarks['test'].points))



from pytest import raises
import numpy as np
from numpy.testing import assert_allclose
from menpo.image import Image
from menpo.feature import gradient

from menpo.feature.features import _np_gradient
import menpo.io as mio


takeo = mio.import_builtin_asset.takeo_ppm()
example_image = np.array([[1., 2., 6.], [3., 4., 5.]])
y_grad = np.array([[2., 2., -1.], [2., 2., -1.]])
x_grad = np.array([[1., 2.5, 4.], [1., 1., 1.]])


def test_gradient_float():
    dtype = np.float32
    p = example_image.astype(dtype)
    image = Image(p)
    grad_image = gradient(image)
    _check_assertions(grad_image, image.shape, image.n_channels * 2,
                      dtype)
    np_grad = np.gradient(p)
    assert_allclose(grad_image.pixels[0], np_grad[0])
    assert_allclose(grad_image.pixels[1], np_grad[1])


def test_gradient_takeo_float32():
    dtype = np.float32
    t = takeo.copy()
    t.pixels = t.pixels.astype(dtype)
    grad_image = gradient(t)
    _check_assertions(grad_image, t.shape, t.n_channels * 2,
                      dtype)
    np_grad = _np_gradient(t.pixels)
    assert_allclose(grad_image.pixels, np_grad)


def test_gradient_double():
    dtype = np.float64
    p = example_image.astype(dtype)
    image = Image(p)
    grad_image = gradient(image)
    _check_assertions(grad_image, image.shape, image.n_channels * 2,
                      dtype)
    np_grad = np.gradient(p)
    assert_allclose(grad_image.pixels[0], np_grad[0])
    assert_allclose(grad_image.pixels[1], np_grad[1])


def test_gradient_takeo_double():
    t = takeo.copy()
    t.pixels = t.pixels.astype(np.float64)
    grad_image = gradient(t)

    np_grad = _np_gradient(t.pixels)
    assert_allclose(grad_image.pixels, np_grad)


def test_gradient_uint8_exception():
    image = Image(example_image.astype(np.uint8))
    with raises(TypeError):
        gradient(image)


def _check_assertions(actual_image, expected_shape, expected_n_channels,
                      expected_type):
    assert (actual_image.pixels.dtype == expected_type)
    assert (type(actual_image) == Image)
    assert (actual_image.shape == expected_shape)
    assert (actual_image.n_channels == expected_n_channels)



import numpy as np
map_coordinates = None  # expensive, from scipy.ndimage
from menpo.transform import Homogeneous

# Store out a transform that simply switches the x and y axis
xy_yx = Homogeneous(np.array([[0., 1., 0.],
                              [1., 0., 0.],
                              [0., 0., 1.]]))


def scipy_interpolation(pixels, points_to_sample, mode='constant', order=1,
                        cval=0.):
    r"""
    Interpolation utilizing scipy map_coordinates function.

    Parameters
    ----------
    pixels : ``(n_channels, M, N, ...)`` `ndarray`
        The image to be sampled from, the first axis containing channel
        information
    points_to_sample : ``(n_points, n_dims)`` `ndarray`
        The points which should be sampled from pixels
    mode : ``{constant, nearest, reflect, wrap}``, optional
        Points outside the boundaries of the input are filled according to the
        given mode
    order : `int,` optional
        The order of the spline interpolation. The order has to be in the
        range [0, 5].
    cval : `float`, optional
        The value that should be used for points that are sampled from
        outside the image bounds if mode is ``constant``.

    Returns
    -------
    sampled_image : `ndarray`
        The pixel information sampled at each of the points.
    """
    global map_coordinates
    if map_coordinates is None:
        from scipy.ndimage import map_coordinates  # expensive
    sampled_pixel_values = []
    # Loop over every channel in image - we know first axis is always channels
    # Note that map_coordinates uses the opposite (dims, points) convention
    # to us so we transpose
    points_to_sample_t = points_to_sample.T
    for i in range(pixels.shape[0]):
        sampled_pixel_values.append(map_coordinates(pixels[i],
                                                    points_to_sample_t,
                                                    mode=mode,
                                                    order=order,
                                                    cval=cval))
    sampled_pixel_values = [v.reshape([1, -1]) for v in sampled_pixel_values]
    return np.concatenate(sampled_pixel_values, axis=0)


try:
    from menpo.external.skimage._warps_cy import _warp_fast

    def cython_interpolation(pixels, template_shape, h_transform, mode='constant',
                             order=1, cval=0.):
        r"""
        Interpolation utilizing skimage fast cython warp function. This method
        assumes that the warp takes the form of a homogeneous transform, and
        thus is much faster for operations such as scaling.

        Parameters
        ----------
        pixels : ``(n_channels, M, N, ...)`` `ndarray`
            The image to be sampled from, the first axis containing channel
            information.
        template_shape : `tuple`
            The shape of the new image that will be sampled
        mode : ``{constant, nearest, reflect, wrap}``, optional
            Points outside the boundaries of the input are filled according to the
            given mode.
        order : int, optional
            The order of the spline interpolation. The order has to be in the
            range [0,5].
        cval : `float`, optional
            The value that should be used for points that are sampled from
            outside the image bounds if mode is 'constant'

        Returns
        -------
        sampled_image : `ndarray`
            The pixel information sampled at each of the points.
        """
        # unfortunately they consider xy -> yx
        matrix = xy_yx.compose_before(h_transform).compose_before(xy_yx).h_matrix
        warped_channels = []
        # Unfortunately, Cython does not seem to support the boolean numpy type,
        # so I think we need to do the cast here. If we don't we lose support
        # for warping BooleanImage.
        # TODO: Can Cython support bool types? If so, skip this
        if pixels.dtype == np.bool:
            in_pixels = pixels.astype(np.uint8)
        else:
            in_pixels = pixels
        # Loop over every channel in image - we know last axis is always channels
        # Note that map_coordinates uses the opposite (dims, points) convention
        # to us so we transpose
        for i in range(pixels.shape[0]):
            warped_channels.append(_warp_fast(in_pixels[i], matrix,
                                              output_shape=template_shape,
                                              mode=mode, order=order, cval=cval))
        warped_channels = [v.reshape([1, -1]) for v in warped_channels]

        result = np.concatenate(warped_channels, axis=0)
        # As above, we need to convert the uint8 back to bool
        if pixels.dtype == np.bool:
            result = result.astype(np.bool)
        return result
except ImportError:
    pass



import os
import subprocess as sp
import warnings

import numpy as np
from pathlib import Path

from menpo.visualize import print_progress
from ..utils import DEVNULL, _call_subprocess


_FFMPEG_CMD = lambda: str(Path(os.environ.get('MENPO_FFMPEG_CMD', 'ffmpeg')))


def ffmpeg_video_exporter(images, out_path, fps=30, codec='libx264',
                          preset='medium', bitrate=None,
                          out_pix_fmt='yuv420p', verbose=False, **kwargs):
    r"""
    Uses subprocess PIPE to export the images using FFMPEG.

    There are is one important environment variable that can be set to alter
    the behaviour of this function:

        ================== ======================================
        ENV Variable       Definition
        ================== ======================================
        MENPO_FFMPEG_CMD   The path to the 'ffmpeg' executable.
        ================== ======================================

    Parameters
    ----------
    images : `list` of :map:`Image`
        List of Menpo images to export as a video.
    out_path : `Path`
        Path to save the video to.
    fps : `int`, optional
        The number of frames per second.
    codec : `str`, optional
        The video codec to use. Default 'libx264', which represents the
        widely available mpeg4. Except when saving .wmv files, then the
        defaults is 'msmpeg4' which is more commonly supported for windows.
    preset : `str`, optional
        The preset FFMPEG compression level.
        Please check out the documentation for more information:
        https://trac.ffmpeg.org/wiki/Encode/H.264#a2.Chooseapreset
    bitrate: `str`, optional
        The output video bitrate.
    out_pix_fmt : `str`, optional
        The output pixel format.
    verbose : `bool`, optional
        If ``True``, print a progress bar.
    **kwargs : `dict`, optional
        Extra parameters for advanced video exporting options.
        They are passed through directly to FFMPEG and they should.
        be organised in pairs of option/value in a dictionary.
        For instance: ``{'crf' : '0'} # equivalent to -crf 0 in ffmpeg.``
        You can find further details in the documentation:
        https://ffmpeg.org/ffmpeg.html#Options
    """
    # Some of the below was inspired by moviepy:
    #   https://github.com/Zulko/moviepy/blob/master/moviepy/video/io/ffmpeg_writer.py
    # and is used under the terms of the MIT license which can be found at
    #   https://github.com/Zulko/moviepy/blob/master/LICENCE.txt
    im = images[0]
    frame_shape = im.shape
    if im.n_channels != 3 and im.n_channels != 1:
        m = ('Currently only images of 1 or 3 channels are expected, '
             'while {} channels were found in the first frame.')
        raise ValueError(m.format(im.n_channels))
    # If the first image is gray then all the images will be assumed to be
    # gray
    colour = 'rgb24' if im.n_channels == 3 else 'gray8'
    cmd = [_FFMPEG_CMD(), '-y',
           '-s', '{}x{}'.format(frame_shape[1], frame_shape[0]),
           '-r', str(fps),
           '-an',
           '-pix_fmt', colour,
           '-c:v', 'rawvideo', '-f', 'rawvideo',
           '-i', '-']
    if codec:
        cmd.extend(['-vcodec', codec])
    if out_pix_fmt:
        cmd.extend(['-pix_fmt', out_pix_fmt])
    if preset:
        cmd.extend(['-preset', preset])
    if bitrate:
        cmd.extend(['-b', str(bitrate)])
    # add the optional kwargs for FFMPEG options.
    for key, value in kwargs.items():
        cmd.extend(['-{}'.format(key), value])
    cmd.append(str(out_path))

    images = (print_progress(images, prefix='Exporting frames') if verbose
              else images)

    # Pipe stdout to DEVNULL to ignore it
    with _call_subprocess(sp.Popen(cmd, stdin=sp.PIPE, stderr=sp.PIPE,
                                   stdout=DEVNULL)) as pipe:
        for k, image in enumerate(images):
            try:
                if image.n_channels != 1 and colour == 'gray8':
                    warnings.warn('Frame {} is non-greyscale and the initial '
                                  'frame was greyscale. This frame will be '
                                  'corrupted.'.format(k))
                if image.shape != frame_shape:  # Valid due to tuple/int
                    warnings.warn('Frame {} is not the same shape as the '
                                  'initial frame and therefore the output '
                                  'may be corrupted.'.format(k))

                i = image.pixels_with_channels_at_back(out_dtype=np.uint8)
                # Handle the case of a greyscale image amidst colour images
                if image.n_channels == 1 and colour == 'rgb24':
                    # Repeat the channels axis 3 times
                    i = i.reshape(i.shape + (1,)).repeat(3, axis=2)
                pipe.stdin.write(i.tostring())
            except IOError:
                error = ('FFMPEG encountered the following error while '
                         'writing the video:\n\n{}'.format(
                    pipe.stderr.read().decode()))
                # Re-raise the error for a useful error message
                raise IOError(error)


def imageio_video_exporter(images, out_path, fps=30, codec='libx264',
                           quality=None, bitrate=None, pixelformat='yuv420p',
                           **kwargs):
    r"""
    Uses imageio to export the images using FFMPEG. Please see the imageio
    documentation for more information.

    Parameters
    ----------
    images : `list` of :map:`Image`
        List of Menpo images to export as a video.
    out_path : `Path`
        Path to save the video to.
    fps : `int`, optional
        The number of frames per second.
    codec : `str`, optional
        The video codec to use. Default 'libx264', which represents the
        widely available mpeg4. Except when saving .wmv files, then the
        defaults is 'msmpeg4' which is more commonly supported for windows
    quality : `float` or `None`
        Video output quality. Uses variable bit rate. Highest
        quality is 10, lowest is 0. Specifying a fixed bitrate using ``bitrate``
        disables this parameter.
    bitrate : `int` or `None`, optional
        Set a constant bitrate for the video encoding. Default is ``None``
        causing ``quality` parameter to be used instead.  Better quality videos
        with smaller file sizes will result from using the ``quality`` variable
        bitrate parameter rather than specifying a fixed bitrate with this
        parameter.
    pixelformat: `str`, optional
        The output video pixel format.
    """
    import imageio

    writer = imageio.get_writer(str(out_path), mode='I', fps=fps,
                                codec=codec, quality=quality, bitrate=bitrate,
                                pixelformat=pixelformat)

    for v in images:
        v = v.pixels_with_channels_at_back(out_dtype=np.uint8)
        writer.append_data(v)
    writer.close()


def image_gif_exporter(images, out_path, fps=30, loop=0, duration=None,
                       **kwargs):
    r"""
    Uses imageio to export the images to a GIF. Please see the imageio
    documentation for more information.

    Parameters
    ----------
    images : `list` of :map:`Image`
        List of Menpo images to export as a video.
    out_path : `Path`
        Path to save the video to.
    fps : `float`, optional
        The number of frames per second. If ``duration`` is not given, the
        duration for each frame is set to 1/fps.
    loop : `int`, optional
        The number of iterations. 0 means loop indefinitely
    duration : `float` or list of `float`, optional
        The duration (in seconds) of each frame. Either specify one value
        that is used for all frames, or one value for each frame.
    """
    import imageio

    writer = imageio.get_writer(str(out_path), mode='I', fps=fps,
                                loop=loop, duration=duration)

    for v in images:
        v = v.pixels_with_channels_at_back(out_dtype=np.uint8)
        writer.append_data(v)
    writer.close()



import numpy as np


def _normalize(v):
    return np.nan_to_num(v / np.sqrt((v ** 2).sum(axis=1, keepdims=True)))


def compute_face_normals(points, trilist):
    """
    Compute per-face normals of the vertices given a list of
    faces.

    Parameters
    ----------
    points : (N, 3) float32/float64 ndarray
        The list of points to compute normals for.
    trilist : (M, 3) int16/int32/int64 ndarray
        The list of faces (triangle list).

    Returns
    -------
    face_normal : (M, 3) float32/float64 ndarray
        The normal per face.
    :return:
    """
    pt = points[trilist]
    a, b, c = pt[:, 0], pt[:, 1], pt[:, 2]
    norm = np.cross(b - a, c - a)
    return _normalize(norm)


def compute_vertex_normals(points, trilist):
    """
    Compute the per-vertex normals of the vertices given a list of
    faces.

    Parameters
    ----------
    points : (N, 3) float32/float64 ndarray
        The list of points to compute normals for.
    trilist : (M, 3) int16/int32/int64 ndarray
        The list of faces (triangle list).

    Returns
    -------
    vertex_normal : (N, 3) float32/float64 ndarray
        The normal per vertex.
    """
    face_normals = compute_face_normals(points, trilist)

    vertex_normals = np.zeros(points.shape, dtype=points.dtype)
    np.add.at(vertex_normals, trilist[:, 0], face_normals)
    np.add.at(vertex_normals, trilist[:, 1], face_normals)
    np.add.at(vertex_normals, trilist[:, 2], face_normals)

    return _normalize(vertex_normals)



import numpy as np
from menpo.transform import Homogeneous, Scale


def tcoords_to_image_coords(image_shape):
    r"""
    Returns a :map:`Homogeneous` transform that converts [0,1]
    texture coordinates (tcoords) used on :map:`TexturedTriMesh`
    instances to image coordinates, which behave just like image landmarks
    do.

    The operations that are performed are:

      - Flipping the origin from bottom-left to top-left
      - Permuting the axis so that  st (or uv) -> yx
      - Scaling the tcoords by the image shape (denormalising them). Note that
        (1, 1) has to map to the highest pixel value, which is actually
        (h - 1, w - 1) due to Menpo being 0-based with image operations.

    Parameters
    ----------
    image_shape : `tuple`
        The shape of the texture that the tcoords index in to.

    Returns
    -------
    :map:`Homogeneous`
        A transform that, when applied to texture coordinates, converts them
        to image coordinates.
    """
    # flip the 'y' st 1 -> 0 and 0 -> 1, moving the axis to upper left
    invert_unit_y = Homogeneous(np.array(
        [[1., 0., 0.],
         [0., -1., 1.],
         [0., 0., 1.]]))

    # flip axis 0 and axis 1 so indexing is as expected
    flip_xy_yx = Homogeneous(np.array(
        [[0., 1., 0.],
         [1., 0., 0.],
         [0., 0., 1.]]))

    return (invert_unit_y
            .compose_before(flip_xy_yx)
            .compose_before(Scale(np.array(image_shape) - 1)))


def image_coords_to_tcoords(image_shape):
    r"""
    Returns a :map:`Homogeneous` transform that converts image coordinates
    (e.g. image landmarks) to texture coordinates (tcoords) as used on
    :map:`TexturedTriMesh` instances.

    The operations that are performed are:

      - Normalizing by image shape (e.g. converting to [0, 1]). Note that
        (1, 1) has to map to the highest pixel value, which is actually
        (h - 1, w - 1) due to Menpo being 0-based with image operations.
      - Permuting the axis so that yx -> st (or uv)
      - Flipping the origin from top-left to bottom-left

    Parameters
    ----------
    image_shape : `tuple`
        The shape of the texture that the image coordinates are on.

    Returns
    -------
    :map:`Homogeneous`
        A transform that, when applied to image coordinates, converts them
        to texture coordinates (tcoords).
    """
    return tcoords_to_image_coords(image_shape).pseudoinverse()



from __future__ import division, print_function
from collections import deque
from datetime import datetime
import sys
from time import time


def progress_bar_str(percentage, bar_length=20, bar_marker='=', show_bar=True):
    r"""
    Returns an `str` of the specified progress percentage. The percentage is
    represented either in the form of a progress bar or in the form of a
    percentage number. It can be combined with the :func:`print_dynamic`
    function.

    Parameters
    ----------
    percentage : `float`
        The progress percentage to be printed. It must be in the range
        ``[0, 1]``.
    bar_length : `int`, optional
        Defines the length of the bar in characters.
    bar_marker : `str`, optional
        Defines the marker character that will be used to fill the bar.
    show_bar : `bool`, optional
        If ``True``, the `str` includes the bar followed by the percentage,
        e.g. ``'[=====     ] 50%'``

        If ``False``, the `str` includes only the percentage,
        e.g. ``'50%'``

    Returns
    -------
    progress_str : `str`
        The progress percentage string that can be printed.

    Raises
    ------
    ValueError
        ``percentage`` is not in the range ``[0, 1]``
    ValueError
        ``bar_length`` must be an integer >= ``1``
    ValueError
        ``bar_marker`` must be a string of length 1

    Examples
    --------
    This for loop: ::

        n_iters = 2000
        for k in range(n_iters):
            print_dynamic(progress_bar_str(float(k) / (n_iters-1)))

    prints a progress bar of the form: ::

        [=============       ] 68%
    """
    if percentage < 0:
        raise ValueError("percentage is not in the range [0, 1]")
    elif percentage > 1:
        percentage = 1
    if not isinstance(bar_length, int) or bar_length < 1:
        raise ValueError("bar_length must be an integer >= 1")
    if not isinstance(bar_marker, str) or len(bar_marker) != 1:
        raise ValueError("bar_marker must be a string of length 1")
    # generate output string
    if show_bar:
        str_param = "[%-" + str(bar_length) + "s] %d%%"
        bar_percentage = int(percentage * bar_length)
        return str_param % (bar_marker * bar_percentage, percentage * 100)
    else:
        return "%d%%" % (percentage * 100)


def print_dynamic(str_to_print):
    r"""
    Prints dynamically the provided `str`, i.e. the `str` is printed and then
    the buffer gets flushed.

    Parameters
    ----------
    str_to_print : `str`
        The string to print.
    """
    sys.stdout.write("\r{}".format(str_to_print.ljust(80)))
    sys.stdout.flush()


def bytes_str(num):
    r"""
    Converts bytes to a human readable format. For example: ::

        print_bytes(12345) returns '12.06 KB'
        print_bytes(123456789) returns '117.74 MB'

    Parameters
    ----------
    num : `int`
        The size in bytes.

    Raises
    ------
    ValueError
        num must be int >= 0
    """
    if not isinstance(num, int) or num < 0:
        raise ValueError("num must be int >= 0")
    for x in ['bytes', 'KB', 'MB', 'GB']:
        if num < 1024.0:
            return "{0:3.2f} {1:s}".format(num, x)
        num /= 1024.0
    return "{0:3.2f} {1:s}".format(num, 'TB')


def print_progress(iterable, prefix='', n_items=None, offset=0,
                   show_bar=True, show_count=True, show_eta=True,
                   end_with_newline=True, min_seconds_between_updates=0.1):
    r"""
    Print the remaining time needed to compute over an iterable.

    To use, wrap an existing iterable with this function before processing in
    a for loop (see example).

    The estimate of the remaining time is based on a moving average of the last
    100 items completed in the loop.

    Parameters
    ----------
    iterable : `iterable`
        An iterable that will be processed. The iterable is passed through by
        this function, with the time taken for each complete iteration logged.
    prefix : `str`, optional
        If provided a string that will be prepended to the progress report at
        each level.
    n_items : `int`, optional
        Allows for ``iterator`` to be a generator whose length will be assumed
        to be `n_items`. If not provided, then ``iterator`` needs to be
        `Sizable`.
    offset : `int`, optional
        Useful in combination with ``n_items`` - report back the progress as
        if `offset` items have already been handled. ``n_items``  will be left
        unchanged.
    show_bar : `bool`, optional
        If False, The progress bar (e.g. [=========      ]) will be hidden.
    show_count : `bool`, optional
        If False, The item count (e.g. (4/25)) will be hidden.
    show_eta : `bool`, optional
        If False, The estimated time to finish (e.g. - 00:00:03 remaining)
        will be hidden.
    end_with_newline : `bool`, optional
        If False, there will be no new line added at the end of the dynamic
        printing. This means the next print statement will overwrite the
        dynamic report presented here. Useful if you want to follow up a
        print_progress with a second print_progress, where the second
        overwrites the first on the same line.
    min_seconds_between_updates : `float`, optional
        The number of seconds that have to pass between two print updates.
        This allows ``print_progress`` to be used on fast iterations without
        incurring a significant overhead. Set to ``0`` to disable this
        throttling.

    Raises
    ------
    ValueError
        ``offset`` provided without ``n_items``

    Examples
    --------
    This for loop: ::

        from time import sleep
        for i in print_progress(range(100)):
            sleep(1)

    prints a progress report of the form: ::

        [=============       ] 70% (7/10) - 00:00:03 remaining
    """
    if n_items is None and offset != 0:
        raise ValueError('offset can only be set when n_items has been'
                         ' manually provided.')
    if prefix != '':
        prefix += ': '
        bar_length = 10
    else:
        bar_length = 20
    n = n_items if n_items is not None else len(iterable)

    timings = deque([], 100)
    time1 = time()
    last_update_time = 0
    for i, x in enumerate(iterable, 1 + offset):
        yield x
        time2 = time()
        timings.append(time2 - time1)
        time1 = time2
        remaining = n - i
        if time2 - last_update_time < min_seconds_between_updates:
            continue
        last_update_time = time2
        duration = datetime.utcfromtimestamp(sum(timings) / len(timings) *
                                             remaining)
        bar_str = progress_bar_str(i / n, bar_length=bar_length,
                                   show_bar=show_bar)
        count_str = ' ({}/{})'.format(i, n) if show_count else ''
        eta_str = (" - {} remaining".format(duration.strftime('%H:%M:%S'))
                   if show_eta else '')
        print_dynamic('{}{}{}{}'.format(prefix, bar_str, count_str, eta_str))

    # the iterable has now finished - to make it clear redraw the progress with
    # a done message. We also hide the eta at this stage.
    count_str = ' ({}/{})'.format(n, n) if show_count else ''
    bar_str = progress_bar_str(1, bar_length=bar_length, show_bar=show_bar)
    print_dynamic('{}{}{} - done.'.format(prefix, bar_str, count_str))

    if end_with_newline:
        print('')



import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal

from menpo.shape import PointCloud, DirectedGraph, UndirectedGraph
from menpo.math import as_matrix

from .. import GMRFModel, GMRFVectorModel


def _compute_sum_cost_block_sparse(samples, test_sample, graph,
                                   n_features_per_vertex, subtract_mean, mode):
    # create ndarray with data
    data = as_matrix(samples, length=None, return_template=False)
    # initialize cost
    cost = 0.
    # for loop over the graph's edges
    for e in graph.edges:
        v1 = e[0]
        v2 = e[1]
        v1_from = v1 * n_features_per_vertex
        v1_to = (v1 + 1) * n_features_per_vertex
        v2_from = v2 * n_features_per_vertex
        v2_to = (v2 + 1) * n_features_per_vertex
        # slice data and test vector
        y = test_sample.as_vector()
        if mode == 'concatenation':
            x = np.hstack((data[:, v1_from:v1_to], data[:, v2_from:v2_to]))
            y = np.hstack((y[v1_from:v1_to], y[v2_from:v2_to]))
        else:
            x = data[:, v1_from:v1_to] - data[:, v2_from:v2_to]
            y = y[v1_from:v1_to] - y[v2_from:v2_to]
        # compute mean and covariance
        cov = np.linalg.inv(np.cov(x.T))
        mean = np.mean(x, axis=0)
        # compute and sum cost
        if subtract_mean:
            v = y - mean
        else:
            v = y
        cost += v.dot(cov).T.dot(v)
    return cost


def _compute_sum_cost_block_diagonal(samples, test_sample, graph,
                                     n_features_per_vertex, subtract_mean):
    # create ndarray with data
    data = as_matrix(samples, length=None, return_template=False)
    # initialize cost
    cost = 0.
    # for loop over the graph's edges
    for v1 in graph.vertices:
        v1_from = v1 * n_features_per_vertex
        v1_to = (v1 + 1) * n_features_per_vertex
        # slice data and test vector
        y = test_sample.as_vector()
        x = data[:, v1_from:v1_to]
        y = y[v1_from:v1_to]
        # compute mean and covariance
        cov = np.linalg.inv(np.cov(x.T))
        mean = np.mean(x, axis=0)
        # compute and sum cost
        if subtract_mean:
            v = y - mean
        else:
            v = y
        cost += v.dot(cov).T.dot(v)
    return cost


def test_mahalanobis_distance():
    # arguments values
    mode_values = ['concatenation', 'subtraction']
    n_features_per_vertex_values = [2, 3]
    sparse_values = [True, False]
    subtract_mean_values = [True, False]
    n_components_values = [None, 30]

    # create graph
    n_vertices = 6
    edges = np.array([[0, 1], [1, 2], [2, 3], [3, 4], [4, 5], [5, 0]])
    graphs = [DirectedGraph.init_from_edges(edges, n_vertices),
              UndirectedGraph(np.zeros((n_vertices, n_vertices)))]

    for n_features_per_vertex in n_features_per_vertex_values:
        # create samples
        n_samples = 50
        samples = []
        for i in range(n_samples):
            samples.append(PointCloud(np.random.rand(n_vertices,
                                                     n_features_per_vertex)))
        test_sample = PointCloud(np.random.rand(n_vertices,
                                                n_features_per_vertex))

        for graph in graphs:
            for mode in mode_values:
                for sparse in sparse_values:
                    for n_components in n_components_values:
                        # train GMRF
                        gmrf = GMRFModel(
                            samples, graph, mode=mode, sparse=sparse,
                            n_components=n_components, dtype=np.float64)

                        for subtract_mean in subtract_mean_values:
                            # compute costs
                            if graph.n_edges == 0:
                                cost1 = _compute_sum_cost_block_diagonal(
                                    samples, test_sample, graph,
                                    n_features_per_vertex, subtract_mean)
                            else:
                                cost1 = _compute_sum_cost_block_sparse(
                                    samples, test_sample, graph,
                                    n_features_per_vertex, subtract_mean, mode)
                            cost2 = gmrf.mahalanobis_distance(
                                test_sample, subtract_mean=subtract_mean)
                            assert_almost_equal(cost1, cost2)


def test_increment():
    # arguments values
    mode_values = ['concatenation', 'subtraction']
    n_features_per_vertex_values = [2, 3]
    sparse_values = [True, False]
    n_components_values = [None, 30]

    # create graph
    n_vertices = 6
    edges = np.array([[0, 1], [1, 2], [2, 3], [3, 4], [4, 5], [5, 0]])
    graphs = [DirectedGraph.init_from_edges(edges, n_vertices),
              UndirectedGraph(np.zeros((n_vertices, n_vertices)))]

    for n_features_per_vertex in n_features_per_vertex_values:
        # create samples
        n_samples = 100
        samples = []
        for i in range(n_samples):
            samples.append(np.random.rand(n_vertices * n_features_per_vertex))

        for graph in graphs:
            for mode in mode_values:
                for sparse in sparse_values:
                    for n_components in n_components_values:
                        # Incremental GMRF
                        gmrf1 = GMRFVectorModel(
                            samples[:50], graph, mode=mode, sparse=sparse,
                            n_components=n_components, dtype=np.float64,
                            incremental=True)
                        gmrf1.increment(samples[50::])

                        # Non incremental GMRF
                        gmrf2 = GMRFVectorModel(
                            samples, graph, mode=mode, sparse=sparse,
                            n_components=n_components, dtype=np.float64)

                        # Compare
                        if sparse:
                            assert_array_almost_equal(gmrf1.precision.todense(),
                                                      gmrf2.precision.todense())
                        assert_array_almost_equal(gmrf1.mean_vector,
                                                  gmrf2.mean_vector)



import collections
import numpy as np

from menpo.shape import TriMesh
from menpo.image import Image
from menpo.compatibility import basestring


def check_param(n, types, param_name, param):
    r"""
    Utility method for validating a given parameter. Since parameters may be
    a single option applied to all items in the list, or a list of different
    options per item, we validate this. If a single argument is passed, it
    is extended into a list.

    Parameters
    ----------
    n : `int`
        Number of items expected.
    types : `type`
        Python `type` or list of `types` that the parameters are expected to be.
    param_name : `str`
        The name of parameter for printing error messages.
    param : `object`
        An object or list of objects that represent a parameter for a
        function.

    Returns
    -------
    param_list : `list` of type `object`
        This is a validated list of length ``n``, one parameter for each
        element in the implicit list of other parameters.
    """
    error_msg = "{0} must be a {1} or a list/tuple of " \
                "{1} with length {2}".format(param_name, types, n)

    # Could be a single value - or we have an error
    if isinstance(param, types):
        return [param] * n
    elif not isinstance(param, collections.Iterable):
        raise ValueError(error_msg)

    # Must be an iterable object
    len_param = len(param)
    isinstance_all_in_param = all(isinstance(p, types) for p in param)

    if len_param == 1 and isinstance_all_in_param:
        return list(param) * n
    elif len_param == n and isinstance_all_in_param:
        return list(param)
    else:
        raise ValueError(error_msg)


def _rasterize_matplotlib(image, pclouds, render_lines=True, line_style='-',
                          line_colour='b', line_width=1, render_markers=True,
                          marker_style='o', marker_size=1,
                          marker_face_colour='b', marker_edge_colour='b',
                          marker_edge_width=1):
    import matplotlib.pyplot as plt
    # TODO: Since upgrading to Matplotlib 2.0 it seems that the rendering acts
    #       strangely with respect to pixel accurate rendering. The lines
    #       appear to be rendering too short - but seem correct when we reset
    #       to the classic style - so obviously there is some matplotlib setting
    #       we could just manually set to get the correct behaviour but it is
    #       a chore to find so I'm pragmatically maintaining the old behaviour
    #       here with this contextmanager.
    with plt.style.context('classic'):
        # Convert image shape into 100 DPI inches
        # This makes sure we maintain the original image size
        image_shape = np.array(image.shape)[::-1] / 100.0
        f = plt.figure(figsize=image_shape, frameon=False, dpi=100)

        image.view(figure_id=f.number, figure_size=image_shape)
        for k, p in enumerate(pclouds):
            p.view(figure_id=f.number, render_axes=False,
                   figure_size=image_shape,
                   render_lines=render_lines[k], line_style=line_style[k],
                   line_colour=line_colour[k], line_width=line_width[k],
                   render_markers=render_markers[k],
                   marker_style=marker_style[k],
                   marker_size=marker_size[k],
                   marker_face_colour=marker_face_colour[k],
                   marker_edge_colour=marker_edge_colour[k],
                   marker_edge_width=marker_edge_width[k])

        # Make sure the layout is tight so that the image is of the original size
        f.tight_layout(pad=0)
        # Get the pixels directly from the canvas buffer which is fast
        c_buffer, shape = f.canvas.print_to_buffer()
        # Turn buffer into numpy array and reshape to image
        shape = shape[::-1] + (-1,)
        pixels_buffer = np.frombuffer(c_buffer, dtype=np.uint8).reshape(shape)
        # Prevent matplotlib from rendering
        plt.close(f)
        # We have to apply the alpha channel to get the correct colour
        mask = pixels_buffer[..., 3:] * (1.0 / 255.0)
        pixels = (pixels_buffer[..., :3] * mask).astype(np.uint8)
        return Image.init_from_channels_at_back(pixels)


def _parse_colour(x):
    if isinstance(x, basestring):
        from matplotlib.colors import ColorConverter
        c = ColorConverter()
        x = c.to_rgb(x)

    if isinstance(x, (tuple, list)):
        # Assume we have a floating point rgb
        if any(a <= 1.0 for a in x):
            x = [int(a * 255) for a in x]
    return tuple(x)


def _rasterize_pillow(image, pclouds, render_lines=True, line_style='-',
                      line_colour='b', line_width=1, render_markers=True,
                      marker_style='o', marker_size=1, marker_face_colour='b',
                      marker_edge_colour='b', marker_edge_width=1):
    from PIL import ImageDraw

    if image.n_channels not in {1, 3}:
        raise ValueError('The Pillow rasterizer only supports grayscale or '
                         'colour images')
    if any(x != '-'for x in line_style):
        raise ValueError("The Pillow rasterizer only supports the '-' "
                         "line style.")
    if any(x not in {'o', 's'} for x in marker_style):
        raise ValueError("The Pillow rasterizer only supports the 'o' and 's' "
                         "marker styles.")
    if any(x > 1 for x in marker_edge_width):
        raise ValueError('The Pillow rasterizer only supports '
                         'marker_edge_width of 1 or 0.')

    if image.n_channels == 1:
        # Make the image RGB
        image = image.extract_channels([0, 0, 0])

    pil_im = image.as_PILImage()
    draw = ImageDraw.Draw(pil_im)

    line_colour = [_parse_colour(x) for x in line_colour]
    marker_edge_colour = [_parse_colour(x) for x in marker_edge_colour]
    marker_face_colour = [_parse_colour(x) for x in marker_face_colour]

    for k in range(len(pclouds)):
        p = pclouds[k]
        if isinstance(p, TriMesh):
            pclouds[k] = p.as_pointgraph()

        points = p.points
        if (render_lines[k] and line_width[k] > 0 and
            hasattr(p, 'edges') and p.edges.size > 0):
            edges = p.edges
            lines = zip(points[edges[:, 0], :],
                        points[edges[:, 1], :])

            for l1, l2 in lines:
                draw.line([tuple(l1[::-1]), tuple(l2[::-1])],
                          fill=line_colour[k], width=line_width[k])

        if render_markers[k] and marker_size[k] > 0:
            draw_func = (draw.ellipse if marker_style[k] == 'o'
                         else draw.rectangle)
            outline = (marker_edge_colour[k] if marker_edge_width[k] == 1
                       else None)
            for p in points:
                y, x = p
                draw_func((x - marker_size[k], y - marker_size[k],
                           x + marker_size[k], y + marker_size[k]),
                          fill=marker_face_colour[k], outline=outline)

    del draw

    pixels = np.asarray(pil_im)
    if image.n_channels == 3:
        return Image.init_from_channels_at_back(pixels)
    else:
        return Image(pixels)


_RASTERIZE_BACKENDS = {'matplotlib': _rasterize_matplotlib,
                       'pillow': _rasterize_pillow}


def rasterize_landmarks_2d(image, group=None, render_lines=True, line_style='-',
                           line_colour='b', line_width=1, render_markers=True,
                           marker_style='o', marker_size=1,
                           marker_face_colour='b', marker_edge_colour='b',
                           marker_edge_width=1, backend='matplotlib'):
    r"""
    This method provides the ability to rasterize 2D landmarks onto an image.
    The returned image has the specified landmark groups rasterized onto
    the image - which is useful for things like creating result examples
    or rendering videos with annotations.

    Since multiple landmark groups can be specified, all arguments can
    take lists of parameters that map to the provided groups list. Therefore,
    the parameters must be lists of the correct length or a single parameter
    to apply to every landmark group.

    Multiple backends are provided, all with different strengths. The 'pillow'
    backend is very fast, but not very flexible. The `matplotlib` backend
    should be feature compatible with other Menpo rendering methods, but
    is much slower due to the overhead of creating a figure to render
    into.

    Parameters
    ----------
    image : :map:`Image` or subclass
        The image to render onto.
    group : `str` or `list` of `str`, optional
        The landmark group key, or a list of keys.
    render_lines : `bool`, optional
        If ``True``, and the provided landmark group is a :map:`PointGraph`,
        the edges are rendered.
    line_style : `str`, optional
        The style of the edge line. Not all backends support this argument.
    line_colour : `str` or `tuple`, optional
        A Matplotlib style colour or a backend dependant colour.
    line_width : `int`, optional
        The width of the line to rasterize.
    render_markers : `bool`, optional
        If ``True``, render markers at the coordinates of each landmark.
    marker_style : `str`, optional
        A Matplotlib marker style. Not all backends support all marker styles.
    marker_size : `int`, optional
        The size of the marker - different backends use different scale
        spaces so consistent output may by difficult.
    marker_face_colour : `str`, optional
        A Matplotlib style colour or a backend dependant colour.
    marker_edge_colour : `str`, optional
        A Matplotlib style colour or a backend dependant colour.
    marker_edge_width : `int`, optional
        The width of the marker edge. Not all backends support this.
    backend : {'matplotlib', 'pillow'}, optional
        The backend to use.

    Returns
    -------
    rasterized_image : :map:`Image`
        The image with the landmarks rasterized directly into the pixels.
        The pixels of the image returned are of uint8 type.

    Raises
    ------
    ValueError
        Only 2D images are supported.
    ValueError
        Only RGB (3-channel) or Greyscale (1-channel) images are supported.
    """
    if image.n_channels != 1 and image.n_channels != 3:
        raise ValueError('Only RGB or Greyscale images can be rasterized')
    if image.n_dims != 2:
        raise ValueError('Only 2D images can be rasterized.')

    if backend in _RASTERIZE_BACKENDS:
        if isinstance(group, list):
            landmarks = [image.landmarks[g] for g in group]
        else:
            landmarks = [image.landmarks[group]]

        # Validate all the parameters for multiple landmark groups being
        # passed in
        n_pclouds = len(landmarks)
        render_lines = check_param(n_pclouds, bool, 'render_lines',
                                   render_lines)
        line_style = check_param(n_pclouds, basestring, 'line_style',
                                 line_style)
        line_colour = check_param(n_pclouds, (basestring, tuple), 'line_colour',
                                  line_colour)
        line_width = check_param(n_pclouds, int, 'line_width', line_width)
        render_markers = check_param(n_pclouds, bool, 'render_markers',
                                     render_markers)
        marker_style = check_param(n_pclouds, basestring, 'marker_style',
                                   marker_style)
        marker_size = check_param(n_pclouds, int, 'marker_size', marker_size)
        marker_face_colour = check_param(n_pclouds, (basestring, tuple),
                                         'marker_face_colour',
                                         marker_face_colour)
        marker_edge_colour = check_param(n_pclouds, (basestring, tuple),
                                         'marker_edge_colour',
                                         marker_edge_colour)
        marker_edge_width = check_param(n_pclouds, int, 'marker_edge_width',
                                        marker_edge_width)

        return _RASTERIZE_BACKENDS[backend](
            image, landmarks, render_lines=render_lines, line_style=line_style,
            line_colour=line_colour, line_width=line_width,
            render_markers=render_markers, marker_style=marker_style,
            marker_size=marker_size, marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width)
    else:
        raise ValueError('Unsupported backend: {}'.format(backend))
